from math import exp, sqrt, log, gamma
from scipy import integrate         # 数値積分関数 integrateを読み込む
import openpyxl


from tklib.tkfile import tkFile
from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me


LorentzNumber_FEA = pi * pi * kB * kB / 3.0 / e / e #2.44e-8 W ohm / K^2

def rieman(x0, dx, y, xmin, xmax):
    S = 0.0
    for i in range(len(y)):
        x = x0 + i * dx
        if x < xmin or xmax < x:
            continue
        S += y[i]
    return S * dx

def integrate_Simpson(func, E0, E1, h, nmin = 8):
    if E0 == E1:
        return 0.0

    n = int((E1 - E0) / h + 1.000001)
    if n < nmin:
        n = nmin
    h = (E1 - E0) / (n - 1)
    y = [func(E0 + i * h) for i in range(n)]
    
    S = 0.5 * (y[0] + y[n-1]) + sum(y[1:n-1])

    return h * S

def integrate_Simpson_list(xlist, ylist):
    n = len(xlist)
    h = xlist[1] - xlist[0]

    S = [0.0]
    for i in range(1, n):
        S.append(S[i-1] + 0.5 * h * (ylist[i] + ylist[i-1]))

    return S

# Fermi-Dirac distribution function
def fe(E, T, EF):
    global e, kB

    if T == 0.0:
        if E < EF:
            return 1.0
        elif E == EF:
            return 0.5
        else:
            return 0.0

    EkT = (E - EF) * e / kB / T
    if EkT > 700.0:
        return 0.0
    elif EkT < -700.0:
        return 1.0

    return 1.0 / (exp(EkT) + 1.0)

def fh(E, T, EF):
    global e, kB

    if T == 0.0:
        if E < EF:
            return 0.0
        elif E == EF:
            return 0.5
        else:
            return 1.0

    EkT = (EF - E) * e / kB / T
    if EkT > 700.0:
        return 0.0
    elif EkT < -700.0:
        return 1.0

    return 1.0 / (exp(EkT) + 1.0)

def DiffFermiDiracDistribution(w, w0, x):
    s = (x - x0) / w
    return exp(s) / w / (1.0 + exp(s)) / (1.0 + exp(s))

# 被積分関数を定義
def func_Fr(x, eta, r):
    if x > 700.0:
        return 0.0
    if x < -700.0:
        return pow(x, r)
    return pow(x, r) / (exp(x - eta) + 1)

# 一番簡単な積分計算
def FermiIntegral_simple(eta, r, eps = 1.0e-7):
    if eta < -13.0:
        return exp(eta) * gamma(r + 1.0)

    xmax = -log(eps*1.0e-1) + abs(r)
    if eta > 0.0:
        ret = integrate.quad(lambda x: func_Fr(x, eta, r), 0.0, eta + xmax)
    else:
        ret = integrate.quad(lambda x: func_Fr(x, eta, r), 0.0, xmax)
    return ret[0]

# 変数変換 z=x^2の場合の非積分関数
def func_Fr_x2(x, eta, r):
    if x > 26.0:
        return 0.0
    if x < -26.0:
        return pow(x*x, r)
    return 2.0 * pow(x, r + r + 1) / (exp(x*x - eta) + 1)

# 変数変換 z=x^2をして積分範囲を短縮、高速化
# 非縮退 ( (EF  EC) / kT < -13) ではΓ関数で近似
def FermiIntegral_x2(eta, r, eps = 1.0e-7):
    if eta < -13.0:
        return exp(eta) * gamma(r + 1.0)
    
    xmax = sqrt(-log(eps*1.0e-1) + abs(r))
    if eta > 0.0:
        ret = integrate.quad(lambda x: func_Fr_x2(x, eta, r), 0.0, eta + xmax)
    else:
        ret = integrate.quad(lambda x: func_Fr_x2(x, eta, r), 0.0, xmax)
    return ret[0]

def FermiIntegral_Simpson(eta, r, dx  = 0.01, eps = 1.0e-7):
    if eta < -13.0:
        return exp(eta) * gamma(r + 1.0)
    
    dx = 0.03
    if eta > 0.0:
        xmax = -log(eps*1.0e-1) + abs(r)
    else:
        xmax = eta - log(eps*1.0e-1) + abs(r)

    nx = int(xmax / dx + 1.000001)
    x = [i * dx for i in range(nx)]
    y = [func_Fr_x2(i * dx, eta, r) for i in range(nx)]
    ret = integrate.simps(y, x, dx)

# Rieman
#    ret = sum(y)
#    ret *= dx

# Trapezoid
#    ret = 0.5 * y[0] + sum(y[1:])
#    ret *= dx

# Simpson
#    ret = sum([4.0 * y[i] + 2.0 * y[i+1] for i in range(1, nx-2, 2)])
#    ret += y[0]
#    ret *= dx / 3.0

# Bode
#    ret = sum([64.0 * y[i] + 24.0 * y[i+1] + 64.0 * y[i+2] + 28.0 * y[i+3] for i in range(1, nx-4, 4)])
#    ret += 14.0 * y[0]
#    ret *= dx / 45.0

    return ret

# degenerate case
DegenerateFD_z = (0.822467, 0.947033, 0.985551, 0.996233, 0.999040,
                  0.999758, 0.999939, 0.999985, 0.999996, 0.999999)
def DegenerateFermiDirac(eta, r, eps = 1.0e-7):
    gi = eta**(r + 1.0) / (r + 1.0)
    t = r
    gg = t / (eta * eta)
    gs = gg * DegenerateFD_z[0]
    t -= 1.0
    for k in range(1, 10):
        gg *= t * (t-1.0) / (eta * eta)
        gs += gg * DegenerateFD_z[k]  
        t -= 2.0;

    return gi * (1.0 + 2.0 * (r + 1.0) * gs)

# non-degenerate case
def NonDegenerateFermiDirac(eta, r, eps = 1.0e-7):
    sum = 0.0;
    s = 1.0;
    while 1:
        gg = exp(s * eta) / s**(r + 1.0)
        sum += gg
        s += 1.0
        gg = exp(s * eta) / s**(r + 1.0)
        sum -= gg
        s += 1.0
        
        if abs(gg) < 1.0e-8:
            break

    return sum * gamma(r + 1.0)

def FermiIntegral_fast(eta, r, eps = 1.0e-7):
    if eta >= 20.0:
        return DegenerateFermiDirac(eta, r, eps)
    if eta < -0.1:
        return NonDegenerateFermiDirac(eta, r, eps)

    lim = 17.0 + 4.5 * r + eta
    ret1 = integrate.quad(lambda x: func_Fr_x2(x, eta, r), 0.0, 0.2)
    ret2 = integrate.quad(lambda x: func_Fr_x2(x, eta, r), 0.2, 1.0)
    ret3 = integrate.quad(lambda x: func_Fr_x2(x, eta, r), 1.0, 5.0)
    ret4 = integrate.quad(lambda x: func_Fr_x2(x, eta, r), 5.0, 10.0)
    ret5 = integrate.quad(lambda x: func_Fr_x2(x, eta, r), 10.0, lim)
    return ret1[0] + ret2[0] + ret3[0] + ret4[0] + ret5[0]

# Effective density of states from effective mass, unit in cm^-3
def meff2NC_FEA(meff, T):
    val = 2.0 * pow(2.0 * pi * me * meff * kB * T / h / h, 1.5) * 1.0e-6
    return val
    
# Effective mass from effective density of states, unit in me
def NC2meff_FEA(NC, T):
    return pow(NC * 1.0e6 / 2.0, 2.0/3.0) / 2.0 / pi / me / kB / T * h * h

# Prefactor of density of states from effective mass, unit in cm^-3eV^-1.5
def meff2DC0_FEA(meff, T):
    return sqrt(2.0) / pi / pi * pow(me * meff, 1.5) / hbar / hbar / hbar * 1.0e-6 * pow(e, 1.5)

def DC02meff_FEA(DC0, T):
    return pow(DC0 / sqrt(2.0) * pi * pi * hbar * hbar * hbar / 1.0e-6 / pow(e, 1.5), 2.0/3.0) / me


# sigma (S/m), S (V/K)
# muw (m^2/(Vs)), nw (m^-3)
def weighted_mobility(sigma, S, T):
    if S < 0.0:
        sign = -1
        carriertype = 'e'
    else:
        sign = 1
        carriertype = 'h'

    absS  = abs(S)
    kBe   = kB / e
    SkBe = absS / kBe

    m2kBT = 2.0 * me * kB * T
    A1b = 8.0 * pi * e * pow(m2kBT, 1.5)
    A1 = 3.0 * h**3 * sigma / A1b
    A2 = 3.0 / (pi * pi) / kBe * absS
    expp1m2 = exp(SkBe - 2.0)
    expm5m1 = exp(-5.0 * (SkBe - 1.0))
    expp5m1 = exp( 5.0 * (SkBe - 1.0))
    
    muw = A1 * (expp1m2 / (1.0 + expm5m1) + A2 / (1.0 + expp5m1))
    nw  = sigma / e / muw

    return muw, nw, sign, carriertype
        
def CalEF_metal(N, T, EF0, dE):
    global D0

    if T == 0.0:
        return pow(1.5 / D0 * N, 2.0 / 3.0)  # eV

    EF = optimize.newton(lambda EF: Ne(T, EF, dE) - N, EF0)
    return EF

def CalNe_metal(T, EF):
    print("T = ", T, "k   EF = ", EF, " eV")
    print("")

    Ianaly = 2.0 / 3.0 * pow(EF, 1.5)
    print("Analytical integration for De(E) from ", 0, " to ", EF, " = ", Ianaly)
    ret = integrate.quad(De, 0.0, EF, epsrel = eps)
    print("quad   : ret=", ret)
    ret = integrate.romberg(De, 0.0, EF, rtol = eps, divmax = maxorder)
    print("romberg: ret=", ret)
    print("")

# 積分範囲
    (Emin, Emax) = (0.0, EF + dE)
    print("(1) Integration range: ", Emin, " - ", Emax, " eV")
    t0 = time.time()
    for i in range(ncycle):
        ret = integrate.quad(lambda E: Defe(E, T, EF), Emin, Emax, epsrel = eps)
    dt = time.time() - t0
    print("quad   : ret=", ret, "  time=", dt, " s")
    t0 = time.time()
    for i in range(ncycle):
        ret = integrate.romberg(lambda E: Defe(E, T, EF), Emin, Emax, rtol = eps, divmax = maxorder)
    dt = time.time() - t0
    print("romberg: ret=", ret, "  time=", dt, " s")
    print("")

# 積分範囲
    (Emin, Emax) = (0.0, EF - dE)
    print("(2) Integration range: ", Emin, " - ", Emax, " eV")
    Ianaly = 2.0 / 3.0 * (pow(Emax, 1.5) - pow(Emin, 1.5))
    print("Analytical integration for De(E) from ", Emin, " to ", Emax, " = ", Ianaly)
    t0 = time.time()
    for i in range(ncycle):
        ret = integrate.quad(lambda E: Defe(E, T, EF), Emin, Emax, epsrel = eps)
    dt = time.time() - t0
    print("quad   : ret=", ret, "  time=", dt, " s")
    t0 = time.time()
    for i in range(ncycle):
        ret = integrate.romberg(lambda E: Defe(E, T, EF), Emin, Emax, rtol = eps, divmax = maxorder)
    dt = time.time() - t0
    print("romberg: ret=", ret, "  time=", dt, " s")
    print("")


# 積分範囲
    (Emin, Emax) = (EF - dE, EF + dE)
    print("(3) Integration range: ", Emin, " - ", Emax, " eV")
    t0 = time.time()
    for i in range(ncycle):
        ret = integrate.quad(lambda E: Defe(E, T, EF), Emin, Emax, epsrel = eps)
    dt = time.time() - t0
    print("quad   : ret=", ret, "  time=", dt, " s")
    t0 = time.time()
    for i in range(ncycle):
        ret = integrate.romberg(lambda E: Defe(E, T, EF), Emin, Emax, rtol = eps, divmax = maxorder)
    dt = time.time() - t0
    print("romberg: ret=", ret, "  time=", dt, " s")
    print("")

def BMShift_FEA(Ne, D0):
    EF = (Ne / D0 * 1.5)**(2.0/3.0)
    return EF

def RichardsonDushman_FEA(meff = 1.0):
    Richardson = 4.0 * pi * me * e * kB * kB / h / h / h
    mef = meff / me
    EffectiveRichardson = Richardson * mef
    return EffectiveRichardson

def CalPlasmaFrequency(Ne, meff):
    return sqrt(Ne * e * e / me / meff / e0) / 2.0 / pi


"""
def cal_mue(self, T):
        t = (T - self.Tnorm) / self.Tnorm
#        t1 = T / self.Tnorm
#        return pow(t1, -0.5) * exp(-self.Ebe * ekB / T) * (self.mue + self.ae1 * t + self.ae2 * t * t)
        return exp(-self.Ebe * ekB / T) * (self.mue + self.ae1 * t + self.ae2 * t * t)

def cal_muh(self, T):
        t = (T - self.Tnorm) / self.Tnorm
#        t1 = T / self.Tnorm
#        return pow(t1, -0.5) * exp(-self.Ebh * ekB / T) * (self.muh + self.ah1 * t + self.ah2 * t * t)
        return exp(-self.Ebh * ekB / T) * (self.muh + self.ah1 * t + self.ah2 * t * t)
"""

def cal_twocarriers(mue, ne, muh, nh):
    muh2mue2 = muh * muh * mue * mue
    nhmuhnemue = nh * muh + ne * mue
    nhmuenemuh = nh * mue + ne * muh
    nmu2diff = nh  * muh * muh - ne * mue * mue

    rhoxx = 1.0 / e / nhmuhnemue

    a4 = nhmuhnemue * nhmuhnemue
    RH = 1 / e * nmu2diff / a4

    return rhoxx, RH, 1.0 / RH / e

def cal_statistics(self, T):
#初期範囲として、価電子帯上端と伝導帯下端エネルギーを設定する
        EFmin = self.Ev - 1.0
        EFmax = self.Ec + 1.0
# まず、EFmin,EFmaxにおけるΔQを計算し、それぞれが正・負あるいは負・生となっていることを確認する    
        dQmin = self.Ne(EFmin, T) + self.NAm(EFmin, T) - self.Nh(EFmin, T) - self.NDp(EFmin, T)
        dQmax = self.Ne(EFmax, T) + self.NAm(EFmax, T) - self.Nh(EFmax, T) - self.NDp(EFmax, T)
        if dQmin * dQmax > 0.0:
            print("Error: Initial Emin and Emax should be chosen as dQmin * dQmax < 0")
            return None, None, None, None, None, None        

# 2分法開始
        for i in range(mat.nmaxiter):
            EFhalf = (EFmin + EFmax) / 2.0
            Neh  = self.Ne(EFhalf, T)
            NAmh = self.NAm(EFhalf, T)
            Nhh  = self.Nh(EFhalf, T)
            NDph = self.NDp(EFhalf, T)
            dQhalf = Neh + NAmh - Nhh - NDph
#        print("  Iter {}: EFhalf = {:12.8f}  dQhalf = {:12.4g}".format(i, EFhalf, dQhalf))
#        print("     Ne={:10.4e}  Nh={:10.4e}  NA-={:10.4e}  ND+={:10.4e}  dQ={:10.4e}".format(Neh, Nhh, NAmh, NDph, dQhalf))
# EFの精度がepsより小さくなったら敬さん終了
            if abs(EFmin - EFhalf) < mat.eps and abs(EFmax - EFhalf) < mat.eps:
#                print("  Success: Convergence reached at EF = {}".format(EFhalf))
                break
            if dQmin * dQhalf < 0.0:
                EFmax = EFhalf
                dQmax = dQhalf
            else:
                EFmin = EFhalf
                dQmin = dQhalf
        else:
            print("  Failed: Convergence did not reach")
            return None

        rhoxx, RH, Ns = cal_twocarriers(self.mue, Neh, self.muh, Nhh)

        return EFhalf, Neh, Nhh, NDph, NAmh, Ns

def cal_transport(self, T):
        EF, Ne, Nh, NDph, NAm, Ns = self.cal_statistics(T)
        muh = self.cal_muh(T)
        mue = self.cal_mue(T)
        sigma = e * (Ne * mue + Nh * muh)
        return EF, Ne, Nh, NDph, NAm, Ns, mue, muh, sigma

def MR_twocarrier(nh, muh, ne,  mue, B):
    global e
    
    B2 = B * B
    muh2mue2 = muh * muh * mue * mue
    nhmuhnemue = nh * muh + ne * mue
    nhmuenemuh = nh * mue + ne * muh
    ndiff = nh - ne
    nmu2diff = nh  * muh * muh - ne * mue * mue

    a1 = nhmuhnemue + nhmuenemuh * muh * mue * B2
    a2 = nhmuhnemue * nhmuhnemue + ndiff * ndiff * muh2mue2 * B2
    rhoxx = 1.0 / e * a1 / a2
    rhoxx0 = 1.0 / e / nhmuhnemue
    MRxx = rhoxx - rhoxx0

    a3a = nmu2diff                      # nh*muh*muh - ne*mue*mue 
    a3b = ndiff * muh2mue2 * B2         # (nh - ne) * muh*muh*mue*mue * B2
    a4a = nhmuhnemue * nhmuhnemue       # pow(nh*muh + ne*mue, 2)
    a4b = ndiff * ndiff * muh2mue2 * B2 #pow(nh - ne, 2) * muh*muh*mue*mue * B2
    a3 = a3a + a3b
    a4 = a4a + a4b
    MRxy = B / e * a3 / a4

    return rhoxx, rhoxx0, MRxx, MRxy

def MR_singlecarrier(n, mu, B):
    global e

    B2 = B * B
    sigmaxx0 = e * n * mu
#    sigmaxx  = sigmaxx0 * (1.0 + mu * mu * B2)
#    rhoxx  = 1.0 / sigmaxx
#    rhoxx0 = 1.0 / sigmaxx0
    rhoxx0 = 1.0 / sigmaxx0
    rhoxx  = rhoxx0 * (1.0 + mu * mu * B2)
    MRxx = rhoxx - rhoxx0

    MRxy = B / e / n

    return rhoxx, rhoxx0, MRxx, MRxy

def MR_samemu(nh, mu, ne, B):
    return MR_twocarrier(nh, mu, ne,  mu, B)


def MR(model, nh, muh, ne,  mue, B):
    if model == 'two':
        return MR_twocarrier(nh, muh, ne,mue, B)
    elif model == 'single':
        return MR_singlecarrier(nh, muh, B)
    elif model == 'same_mu':
        return MR_samemu(nh, muh, ne, B)


#=============================
# other functions
#=============================
def read_datafile(infile, usage = None):
    datafile = tkVariousData(infile)
    labels, datalist = datafile.Read_minimum_matrix(close_fp = True, usage = usage)
#    print("labels: ", labels)
#    ncol  = len(datalist)
#    print("ncol: ", ncol)
    label_sample, xsample = datafile.FindDataArray(r'^[^(S|sigma|N|mu)][$|\S]?.*$')
#    print("l=", label_sample, xsample)
    label_S, yS           = datafile.FindDataArray(r'^S[$|\S]?.*$')
#    print("l=", label_S, yS)
    label_sigma, ysigma   = datafile.FindDataArray(r'^sigma[$|\S]?.*$', flag = 'i')
    label_N, yN           = datafile.FindDataArray(r'^N[$|\S]?.*$', flag = 'i')
    label_mu, ymu         = datafile.FindDataArray(r'^mu[$|\S]?.*$', flag = 'i')
    return label_sample, xsample, label_S, yS, label_sigma, ysigma, label_N, yN, label_mu, ymu

def read_Hall_excel(infile):
        wb = openpyxl.load_workbook(infile, data_only = True)
        if not wb:
            print("")
            print("Error to read [{}]".format(infile))
            print("")
            terminate()

        sheetnames = wb.sheetnames
        print("sheet names:", sheetnames)
        print("  read [{}]".format(sheetnames[0]))

#       ws = wb.active
        ws = wb[sheetnames[0]]
        Tlabel  = ws[1][0].value
        Nlabel  = ws[1][1].value
        mulabel = ws[1][2].value
        slabel  = ws[1][3].value
        print("Labels:", Tlabel, Nlabel, slabel, mulabel)
        xT  = []
        yN  = []
        ymu = []
        ys  = []
        i = 2
        while 1:
            Tcell = ws.cell(row = i, column = 1).value
            if Tcell is None or Tcell == '':
                break

            Ncell  = ws.cell(row = i, column = 2).value
            mucell = ws.cell(row = i, column = 3).value
            scell  = ws.cell(row = i, column = 4).value
#            print("cells=", Tcell, Ncell, scell, mucell)

            Tcell  = float(Tcell)
            Ncell  = float(Ncell)
            scell  = float(scell)
            mucell = float(mucell)

            xT.append(Tcell)
            yN.append(Ncell)
            ys.append(scell)
            ymu.append(mucell)
            i += 1

        return [Tlabel, Nlabel, mulabel, slabel], xT, yN, ymu, ys
    
def savecsv(outfile, header, datalist):
    try: 
        print("Write to [{}]".format(outfile))
        f = open(outfile, 'w')
    except:
#    except IOError:
        print("Error: Can not write to [{}]".format(outfile))
    else:
        fout = csv.writer(f, lineterminator='\n')
        fout.writerow(header)
#        fout.writerows(data)
        for i in range(0, len(datalist[0])):
            a = []
            for j in range(len(datalist)):
                a.append(datalist[j][i])
            fout.writerow(a)
        f.close()

def read_csv(infile, xmin = None, xmax = None, delimiter = ','):
    print("xrange=", xmin, xmax)
    data = []
    try:
        infp = open(infile, "r")
        f = csv.reader(infp, delimiter = delimiter)
        header = next(f)
        print("header=", header)
        for j in range(len(header)):
            data.append([])

        for row in f:
            x = pfloat(row[0])
            if xmin is not None and xmin <= x <= xmax:
                y = pfloat(row[1])
                data[0].append(x)
                data[1].append(y)
    except:
        print("Error: Can not read [{}]".format(infile))
        exit()
    return header, data[0], data[1]

def read_excel(infile):
    wb = openpyxl.load_workbook(infile, data_only = True)
    if not wb:
        print("")
        print("Error to read [{}]".format(infile))
        print("")
        terminate()

    sheetnames = wb.sheetnames
    print("sheet names:", sheetnames)
    print("  read [{}]".format(sheetnames[0]))

#    ws = wb.active
    ws = wb[sheetnames[0]]
    Tlabel = ws[1][0].value
    Nlabel = ws[1][1].value
    Slabel = ws[1][2].value
    print("Labels:", Tlabel, Nlabel, Slabel)
    xT    = []
    yN    = []
    yS    = []
    i = 2
    while 1:
        Tcell = ws.cell(row = i, column = 1).value
        if Tcell is None or Tcell == '':
            break
        Ncell = ws.cell(row = i, column = 2).value
        Scell = ws.cell(row = i, column = 3).value

        Tcell = float(Tcell)
        Ncell = float(Ncell)
        Scell = float(Scell)

        xT.append(Tcell)
        yN.append(Ncell)
        yS.append(Scell)
        i += 1

    return [Tlabel, Nlabel], xT, yN, yS


def main():
    print("")
    print("This is a library, not runnable")
    print("")


if __name__ == "__main__":
    main()
