import sys
import numpy as np
import openpyxl
import pandas as pd
import matplotlib.pyplot as plt


from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg, analyze_varstr, save_csv
from tklib.tkutils import is_exist, is_file, is_dir
from tklib.tkapplication import tkApplication
from tklib.tkparams import tkParams

from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tkmatrix import make_matrix1
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me, Ea_Arrhenius, log10
from tklib.tksci.tkoptimize import mlsq_general
from tklib.tktransport.tkTransport import integrate_Simpson_list, fe, fh, NC2meff_FEA
from tklib.tktransport.tkmobility import tkMobility



norder = 3
infile  =  'random-poly.xlsx'
xmin = -1.0e100
xmax =  1.0e100
xcalmin = '*'
xcalmax = '*'
xlabel = 0
ylabel = 1

fontsize = 16

app = tkApplication()

argv = sys.argv
narg = len(argv)
if narg >= 2:
    infile = argv[1]
if narg >= 3:
    norder = pint(argv[2])
if narg >= 4:
    xlabel = argv[3]
if narg >= 5:
    ylabel = argv[4]
if narg >= 6:
    xmin = pfloat(argv[5])
if narg >= 7:
    xmax = pfloat(argv[6])
if narg >= 8:
    xcalmin = pfloat(argv[7], defval = xcalmin)
if narg >= 9:
    xcalmax = pfloat(argv[8], defval = xcalmax)


def main():
    global xlabel, ylabel, xcalmin, xcalmax

    logfile = app.replace_path(infile)
    print(f"Open logfile [{logfile}]")
    app.redict(targets = ["stdout", logfile], mode = 'w')

    print("Least-squares method for polynomial order {}".format(norder))
    print(f"infile: {infile}")
    print(f"xlabel: {xlabel}")
    print(f"ylabel: {ylabel}")
    print(f"norder: {norder}")
    print(f"Fitting data range: {xmin} - {xmax}")
    if xlabel == '':
        app.terminate("\nError: Null xlabel is not allowed\n", pause = True)
    if ylabel == '':
        app.terminate("\nError: Null ylabel is not allowed\n", pause = True)

    print("")
    print(f"Read [{infile}]")
    datafile = tkVariousData(infile)
    labels, datalist = datafile.Read_minimum_matrix(close_fp = True, force_numeric = False)
    _xlabel, xin = datafile.FindDataArray(xlabel, flag = 'i')
    _ylabel, yin = datafile.FindDataArray(ylabel, flag = 'i')
    if xin is None:
        app.terminate("\nError: xlabel [{xlabel} is not found\n", pause = True)
    if yin is None:
        app.terminate("\nError: ylabel [{ylabel} is not found\n", pause = True)

    x = []
    y = []
    for i in range(len(xin)):
        if xmin <= xin[i] <= xmax:
            x.append(xin[i])
            y.append(yin[i])
    ndata = len(x)
    print("")
    print("ndata=", ndata)
    print(f"{labels[0]:12} {labels[1]:12}")
    for i in range(ndata):
        print(f"{x[i]:12.4g} {y[i]:12.4g}")

    xcalmin = pfloat(xcalmin, defval = min(x))
    xcalmax = pfloat(xcalmax, defval = max(x))
    ncaldata = 201
    xcalstep = (xcalmax - xcalmin) / (ncaldata - 1)
    xcal = [xcalmin + i * xcalstep for i in range(ncaldata)]
    print(f"Calculation range: {xcalmin} - {xcalmax}")
    print(f"  Number of calculation points: {ncaldata}")

    print("")
    print(f"Execute linear least-squares method")
    ci = np.polyfit(x, y, norder)

    print("LSQ function")
    print("f(x) = {}".format(ci[norder]), end = '')
    for i in range(1, norder+1):
        print(" + {} * x^{}".format(ci[norder-i], i), end = '')
    print("")


    ycal =  np.poly1d(ci)(x)
    print("")
    print("Fitting result")
    print(f"{xlabel:12} {ylabel:12} {ylabel+'(fit)':12}")
    for i in range(ndata):
        print(f"{x[i]:12.4g} {y[i]:12.4g} {ycal[i]:12.4g}")
    
    outfile = app.replace_path(infile, template = ["{dirname}", "{filebody}-fit.xlsx"])
    print("")
    print(f"Save results to [{outfile}]")
    df = pd.DataFrame(np.array([x, y, ycal]).T, columns = [xlabel, ylabel, ylabel + '(fit)'])
    df.to_excel(outfile, index = False)

    ycal =  np.poly1d(ci)(xcal)
    outfile = app.replace_path(infile, template = ["{dirname}", "{filebody}-cal.xlsx"])
    print("")
    print(f"Save results to [{outfile}]")
    df = pd.DataFrame(np.array([xcal, ycal]).T, columns = [xlabel, ylabel + '(cal)'])
    df.to_excel(outfile, index = False)


    """
# poly1dでこの部分の計算をしてくれている
    ycal = []
    for i in range(ndata):
        _x = x[i]
        yl = 0.0
        for k in range(norder+1):
            yl += ci[k] * pow(_x, norder - k)

        ycal.append(yl)
    """

#================================================================
# Plot
#================================================================
    fig, axes = plt.subplots(1, 2, figsize = (8, 6))
    axes[0].plot(xin, yin,    label = 'input', linestyle = '', marker = 'o')
    axes[0].plot(xcal, ycal, label = 'fit',   linestyle = '-')
#    axes[0].plot(x, ycal, label = 'fit',   linestyle = '-')
    axes[0].tick_params(labelsize = fontsize)
    axes[0].set_xlabel(xlabel, fontsize = fontsize)
    axes[0].set_ylabel(ylabel, fontsize = fontsize)
    axes[0].legend(fontsize = fontsize)

    axes[1].plot(range(norder+1), ci, label = 'coeff', linestyle = '', marker = 'o')
    axes[1].tick_params(labelsize = fontsize)
    axes[1].set_xlabel('$i$', fontsize = fontsize)
    axes[1].set_ylabel('$c_i$', fontsize = fontsize)
    axes[1].legend(fontsize = fontsize)

    plt.tight_layout()
    plt.pause(0.1)

    app.terminate("", pause = True)


if __name__ == "__main__":
    main()
